mod readmap_helper;
use crossterm::event::KeyCode;
use readmap::types::MemoryMap;
use std::path::PathBuf;
use tool_helper::Error;
use tui::{
    layout::{Alignment, Constraint, Direction, Layout, Rect},
    style::{Color, Modifier, Style},
    text::{Span, Spans},
    widgets::{Block, Borders, BorderType, Gauge, Paragraph, Tabs}
};

pub type EventReceiver = std::sync::mpsc::Receiver<Event<crossterm::event::KeyEvent>>;
pub type Terminal      = tui::Terminal<tui::backend::CrosstermBackend<std::io::Stdout>>;
type ConsoleFrame<'a>  = tui::Frame<'a, tui::backend::CrosstermBackend<std::io::Stdout>>;

pub enum Event<I> {
    Input(I),
    Tick,
    ForceRender,
}

pub enum UIState {
    Alive,
    Render,
    Terminated
}

pub enum MenuSelection {
    Stats,
    Quit,
}

impl MenuSelection {
    fn next(&self) -> MenuSelection {
        match self {
            Self::Stats => Self::Quit,
            Self::Quit  => Self::Stats
        }
    }

    fn prev(&self) -> MenuSelection {
        self.next()
    }

    fn get_id(&self) -> usize {
        match self {
            MenuSelection::Stats => 0,
            MenuSelection::Quit  => 1,
        }
    }
}

pub struct ConsoleUI {
    rx:             EventReceiver,
    terminal:       Terminal,
    cur_size:       Rect,
    main_chunks:    Vec<Rect>,
    stats_chunks:   Vec<Rect>,
    menu_selection: MenuSelection,
}

impl ConsoleUI {
    pub fn new(rx: EventReceiver, terminal: Terminal) -> ConsoleUI {
        ConsoleUI{rx, terminal, cur_size: Rect::default(), main_chunks: Vec::default(), stats_chunks: Vec::default(), menu_selection: MenuSelection::Stats}
    }

    pub fn update(&mut self) -> Result<UIState, Error> {
        match self.rx.recv()? {
            Event::Input(event) => {
                match event.code {
                    KeyCode::Char('q') => Ok(UIState::Terminated),
                    KeyCode::Char('s') => {
                        self.menu_selection = MenuSelection::Stats;
                        Ok(UIState::Render)
                    },
                    KeyCode::Left      => {
                        self.menu_selection = self.menu_selection.prev();
                        Ok(UIState::Render)
                    },
                    KeyCode::Right     => {
                        self.menu_selection = self.menu_selection.next();
                        Ok(UIState::Render)
                    },
                    KeyCode::Enter     => {
                        if matches!(self.menu_selection, MenuSelection::Quit) {
                            Ok(UIState::Terminated)
                        }
                        else {
                            Ok(UIState::Alive)
                        }
                    },
                    _                  => Ok(UIState::Alive)
                }
            },
            Event::Tick        => {
                if self.terminal.size().expect("Getting size of terminal") != self.cur_size {
                    Ok(UIState::Render)
                } 

                else {
                    Ok(UIState::Alive)
                }
            },
            Event::ForceRender => Ok(UIState::Render)
        }
    }

    pub fn render(&mut self, data: &ConsoleUIData) -> Result<(), Error> {
        self.terminal.draw(|frame| {
            Self::update_sub_frames(&mut self.main_chunks, &mut self.stats_chunks, &mut self.cur_size, frame.size());

            frame.render_widget(Self::create_titel(),                    self.main_chunks[0]);
            frame.render_widget(Self::create_menu(&self.menu_selection), self.main_chunks[1]);
            match self.menu_selection {
                MenuSelection::Stats => {
                    Self::render_stats(frame, &self.stats_chunks, data);
                },
                MenuSelection::Quit  => {
                    frame.render_widget(Self::create_exit_message(), self.main_chunks[2]);
                }
            }
        })?;
        Ok(())
    }

    pub fn close(mut self) -> Result<(), Error> {
        crossterm::terminal::disable_raw_mode()?;
        self.terminal.show_cursor()?;
        self.terminal.clear()?;

        Ok(())
    }

    fn update_sub_frames(sub_frames: &mut Vec<Rect>, stats_frames: &mut Vec<Rect>, frame: &mut Rect, new_frame: Rect) {
        if new_frame  != *frame {
            *sub_frames = Layout::default().direction(Direction::Vertical).constraints([
                Constraint::Length(3),
                Constraint::Length(3),
                Constraint::Min(3)
            ]).split(new_frame);

            *stats_frames = Layout::default().direction(Direction::Vertical).constraints([
                Constraint::Min(3),
                Constraint::Length(3)
            ]).split(sub_frames[2]);

            *frame = new_frame;
        }
    }

    fn create_titel<'a>() -> Paragraph<'a> {
        Paragraph::new("psxreadmap")
            .style(Style::default().fg(Color::White))
            .alignment(Alignment::Center)
            .block(Block::default()
                .borders(Borders::ALL)
                .style(Style::default().fg(Color::White))
                .border_type(BorderType::Plain)
            )
    }

    fn create_menu<'a>(menu_selection: &MenuSelection) -> Tabs<'a> {
        const MENU_TITLES: &'static [&'static str] = &["Stats", "Quit"];

        let menu = MENU_TITLES.iter().map(|t| {
            let (first, rest) = t.split_at(1);
            Spans::from(vec![
                Span::styled(first, Style::default().fg(Color::Yellow).add_modifier(Modifier::UNDERLINED),),
                Span::styled(rest,  Style::default().fg(Color::White)),
            ])
        }).collect();
        
        Tabs::new(menu).select(menu_selection.get_id()).block(
            Block::default()
            .title("Menu").borders(Borders::ALL))
            .style(Style::default().fg(Color::White))
            .highlight_style(Style::default().bg(Color::LightYellow))
            .divider(Span::raw("|")
        )
    }

    fn render_stats(frame: &mut ConsoleFrame, frames: &Vec<Rect>, data: &ConsoleUIData) {
        let content_text = Paragraph::new("<EMPTY>")
            .style(Style::default().fg(Color::White))
            .alignment(Alignment::Center)
            .block(Block::default()
                .borders(Borders::ALL)
                .style(Style::default().fg(Color::White))
                .border_type(BorderType::Plain)
            );
        let mem_gauge = Self::create_overall_mem_gauge(data.highest_adr);

        frame.render_widget(content_text, frames[0]);
        frame.render_widget(mem_gauge,    frames[1]);
    }

    fn create_overall_mem_gauge<'a>(highest_adr: u64) -> Gauge<'a> {
        const RAM_TOP:u64 = ConsoleUIData::HIGHEST_RAM_ADDRESS & !ConsoleUIData::RAM_BASE_ADDRESS;

        let highest_adr_masked = highest_adr & !ConsoleUIData::RAM_BASE_ADDRESS;
        let adr_ratio          = (highest_adr_masked as f64)/RAM_TOP as f64;

        Gauge::default().block(Block::default().borders(Borders::ALL)
        .title(format!("Memory Usage: {}% [0x{:x}/0x{:x}]", adr_ratio*100.0f64, highest_adr, ConsoleUIData::HIGHEST_RAM_ADDRESS)))
        .gauge_style(Style::default().fg(Color::Cyan).bg(Color::Black).add_modifier(Modifier::BOLD))
        .ratio(adr_ratio)
    }

    fn create_exit_message<'a>() -> Paragraph<'a> {
        Paragraph::new("Press \"ENTER\" to exit")
            .style(Style::default().fg(Color::White))
            .alignment(Alignment::Center)
            .block(Block::default()
                .borders(Borders::ALL)
                .style(Style::default().fg(Color::White))
                .border_type(BorderType::Plain).title("QUIT")
            )
    }
}

#[derive(Default)]
pub struct ConsoleUIData {
    highest_adr: u64
}

impl ConsoleUIData {
    pub const RAM_BASE_ADDRESS:u64    = 0x80000000;
    pub const HIGHEST_RAM_ADDRESS:u64 = Self::RAM_BASE_ADDRESS + (2*1024*1024);

    pub fn new(memory_map: &MemoryMap) -> Result<ConsoleUIData, Error> {
        let mut ui_data = ConsoleUIData::default();

        ui_data.update(memory_map)?;
        Ok(ui_data)
    }

    pub fn update(&mut self, memory_map: &MemoryMap) -> Result<(), Error> {
        fn get_last_section_end_adr(memory_map: &MemoryMap) -> u64 {
            if let Some(section) = memory_map.sections.last() {
                return section.adr + section.size as u64;
            }

            0u64
        }

        self.highest_adr = get_last_section_end_adr(memory_map);
        Ok(())
    }
}

pub fn load_memory_map(use_wsl: bool, input: PathBuf) -> Result<MemoryMap, Error> {
    readmap_helper::generate_memory_map(use_wsl, input)
}